@@ -5,18 +5,23 @@ Note that importing this module has a basic semantic: it prepares the dataClay
 core and sets the "client" mode for the library.
 """
 import logging.config
-import os.path
+import os
 import sys
+import warnings
+
 
 from dataclay import getRuntime
 from dataclay.DataClayObject import DataClayObject
-from dataclay.commonruntime.ClientRuntime import *
+from dataclay.commonruntime.ClientRuntime import settings, LANG_PYTHON
 from dataclay.commonruntime.ClientRuntime import UNDEFINED_LOCAL as _UNDEFINED_LOCAL
 from dataclay.commonruntime.Initializer import initialize, _get_logging_dict_config
 from dataclay.communication.grpc.clients.LogicModuleGrpcClient import LMClient
 from dataclay.paraver import TRACE_ENABLED, extrae_tracing_is_enabled, get_task_id, \
     set_current_available_task_id
 from dataclay.util.StubUtils import track_local_available_classes
+from dataclay.util.StubUtils import clean_babel_data
+from dataclay.commonruntime.Settings import unload_settings
+
 from time import sleep
 
 # This will be populated during initialization
@@ -29,7 +34,7 @@ _connection_initialized = False
 _initialized = False
 
 
-def is_initialized():
+def is_initialized() -> bool:
     """Simple query for the _initialized flag.
 
     :return: True if `init` has already been called, False otherwise.
@@ -37,16 +42,24 @@ def is_initialized():
     return _initialized
 
 
-def reinitialize_logging():
+def reinitialize_logging() -> None:
+    """
+    Restart logging system with new logging dict configuration
+    :return: None
+    """
+    warnings.warn("deprecated", DeprecationWarning)
     dictconfig = _get_logging_dict_config()
-
     logger.debug("Ready to close loggers, bye bye!")
     dictconfig["disable_existing_loggers"] = False
     logging.config.dictConfig(dictconfig)
     logger.verbose("Logging reinitialized. Welcome back!")
 
 
-def reinitialize_clients():
+def reinitialize_clients() -> None:
+    """
+    Reinitialize connection to logic module
+    :return: None
+    """
     runtime = getRuntime()
     logger.verbose("Performing reinitialization of clients, removing #%d cached ones and recreating LMClient",
                 len(runtime.ready_clients))
@@ -55,7 +68,7 @@ def reinitialize_clients():
     }
 
 
-def init_connection(client_file):
+def init_connection(client_file) -> LMClient:
     """Initialize the connection client ==> LogicModule.
 
     Note that the connection can be initialized standalone from here (like the
@@ -89,7 +102,7 @@ def init_connection(client_file):
     runtime.ready_clients[settings.logicmodule_dc_instance_id] = runtime.ready_clients["@LM"]
 
     # wait for 1 python backend 
-    while len(get_backends()) < 1:
+    while len(get_backends_info()) < 1:
         logger.info("Waiting for any python backend to be ready ...")
         sleep(2)
 
@@ -98,24 +111,40 @@ def init_connection(client_file):
 
 def get_backends():
     """Return all the dataClay backend present in the system."""
-    result = getRuntime().get_execution_environments_names("admin", "admin")
+    result = getRuntime().get_execution_environments_names(force_update=True)
     logger.debug("Got %i python backend/s", len(result))
     return result
 
 
 def get_backends_info():
     """Return all the dataClay BackendInfo present in the system."""
-    result = getRuntime().get_execution_environments_info()
+    result = getRuntime().get_all_execution_environments_info(force_update=True)
     logger.debug("Got %i python backend/s", len(result))
     return result
 
 
 def get_backend_id_by_name(name):
     """Return dataClay backend present in the system with name provided."""
-    all_backends = getRuntime().get_execution_environments_info()
+    all_backends = getRuntime().get_all_execution_environments_with_name(name)
     for backend in all_backends.values():
         if backend.name == name:
-            return backend.dataClayID
+            return backend.id
+    return None
+
+def get_external_backend_id_by_name(name, external_dataclay_id):
+    """Return dataClay backend present in the system with name provided."""
+    all_backends = getRuntime().get_all_execution_environments_at_dataclay(external_dataclay_id)
+    for backend in all_backends.values():
+        if backend.name == name:
+            return backend.id
+    return None
+
+def get_backend_id(hostname, port):
+    """Return dataClay backend present in the system with name provided."""
+    all_backends = getRuntime().get_all_execution_environments_info(force_update=True)
+    for backend in all_backends.values():
+        if backend.hostname == hostname and backend.port == port:
+            return backend.id
     return None
 
 
@@ -142,6 +171,17 @@ def get_dataclay_id(exthostname, extport):
     """
     return getRuntime().get_external_dataclay_id(exthostname, extport)
 
+def import_models_from_external_dataclay(namespace, ext_dataclay_id) -> None:
+    """ Import models in namespace specified from an external dataClay
+    :param namespace: external dataClay namespace to get
+    :param ext_dataclay_id: external dataClay ID
+    :return: None
+    :type namespace: string
+    :type ext_dataclay_id: UUID
+    :rtype: None
+    """
+    return getRuntime().import_models_from_external_dataclay(namespace, ext_dataclay_id)
+
 
 def unfederate(ext_dataclay_id=None):
     """ Unfederate all objects belonging to/federated with external data clay with id provided 
@@ -175,13 +215,19 @@ def federate_all_objects(dest_dataclay_id):
     """
     return getRuntime().federate_all_objects(dest_dataclay_id)
 
+def get_num_objects():
+    """ Get number of objects in dataClay
+    :return: number of objects in dataClay
+    :rtype: int32
+    """
+    return getRuntime().get_num_objects()
 
 def pre_network_init(config_file):
     """Perform a partial initialization, with no network."""
     settings.load_properties(config_file)
 
 
-def init(config_file=None):
+def init(config_file=None) -> None:
     """Initialization made on the client-side, with file-based settings.
 
     Note that after a successful call to this method, subsequent calls will be
@@ -248,10 +294,10 @@ def post_network_init():
         LANG_PYTHON
     )
     settings.current_session_id = session_info.sessionID
-
+    logger.debug(f"Started session {settings.current_session_id}")
     name = settings.local_backend_name
     if name:
-        exec_envs = getRuntime().get_execution_environments_info()
+        exec_envs = getRuntime().get_all_execution_environments_info()
         for k, v in exec_envs.items():
             if exec_envs[k].name == name:
                 global LOCAL
@@ -302,7 +348,7 @@ def post_network_init():
 
 def finish_tracing():
     """
-    Finishes tracing if needed 
+    Finishes tracing if needed
     """
     if extrae_tracing_is_enabled():
         extrae_compss = int(settings.extrae_starting_task_id) != 0
@@ -319,15 +365,36 @@ def finish_tracing():
                 getRuntime().deactivate_tracing_in_dataclay_services()
                 getRuntime().deactivate_tracing(True)
                 getRuntime().get_traces_in_dataclay_services()  # not on workers!
+                # Merge
+                os.system("mpi2prv -keep-mpits -no-syn -f TRACE.mpits -o ./trace/dctrace.prv")
             else:
                 getRuntime().deactivate_tracing(True)
 
+
+
             
 def finish():
     global _initialized
+    if not _initialized:
+        logger.warning("Already finished --ignoring")
+        return
+    global _connection_initialized
     logger.info("Finishing dataClay API")
     finish_tracing()
+    getRuntime().close_session()
+    logger.debug(f"Closed session {settings.current_session_id}")
     getRuntime().stop_runtime()
+    # Unload stubs
+    clean_babel_data()
+    sys.path.remove(os.path.join(settings.stubs_folder, 'sources'))
+    # unload caches of stubs
+    from dataclay.commonruntime.ExecutionGateway import loaded_classes, class_extradata_cache_client, \
+        class_extradata_cache_exec_env
+    loaded_classes.clear()
+    class_extradata_cache_exec_env.clear()
+    class_extradata_cache_client.clear()
+    # unload settings
+    unload_settings()
     _initialized = False
     _connection_initialized = False
 
