@@ -14,6 +14,9 @@ from dataclay.heap.ClientHeapManager import ClientHeapManager
 from dataclay.loader.ClientObjectLoader import ClientObjectLoader
 
 # Sentinel-like object to catch some typical mistakes
+from dataclay.util.management.metadataservice.MetaDataInfo import MetaDataInfo
+from dataclay.util.management.metadataservice.RegistrationInfo import RegistrationInfo
+import traceback
 UNDEFINED_LOCAL = object()
 
 
@@ -51,7 +54,8 @@ class ClientRuntime(DataClayRuntime):
         :rtype: DataClayID
         :raises RuntimeError: if backend id is UNDEFINED_LOCAL.
         """
-        self.logger.debug("Starting make persistent object for instance %s with id %s", instance,
+
+        self.logger.debug("Starting make persistent object for instance with id %s",
                      instance.get_object_id())
         if backend_id is UNDEFINED_LOCAL:
             # This is a commonruntime end user pitfall,
@@ -88,10 +92,13 @@ class ClientRuntime(DataClayRuntime):
 
                 # From client side, we cannot check if object is registered or not (we do not have isPendingToRegister like EE)
                 # Therefore, we call LogicModule with all information for registration.
-                reg_info = [instance.get_object_id(), instance.get_class_extradata().class_id,
-                            self.get_session_id(), instance.get_dataset_id()]
-                new_object_id = self.ready_clients["@LM"].register_object(reg_info, location, alias, LANG_PYTHON)
-                
+                reg_infos = list()
+                reg_info = RegistrationInfo(instance.get_object_id(), instance.get_class_extradata().class_id,
+                            self.get_session_id(), instance.get_dataset_id(), alias)
+                reg_infos.append(reg_info)
+                new_object_ids = self.ready_clients["@LM"].register_objects(reg_infos, location, LANG_PYTHON)
+                self.logger.debug(f"Received ids: {new_object_ids}")
+                new_object_id = next(iter(new_object_ids))
                 self.update_object_id(instance, new_object_id)
                 
                 self.alias_cache[alias] = instance.get_object_id(), instance.get_class_extradata().class_id, location
@@ -100,6 +107,7 @@ class ClientRuntime(DataClayRuntime):
             self.logger.debug("Instance with object ID %s being send to EE", instance.get_object_id())
             # set the default master location
             instance.set_master_location(location)
+            instance.set_alias(alias)
             # We serialize objects like volatile parameters
             parameters = list()
             parameters.append(instance)
@@ -118,13 +126,14 @@ class ClientRuntime(DataClayRuntime):
             
             # Avoid some race-conditions in communication (make persistent + execute where
             # execute arrives before).
-            self.add_volatiles_under_deserialization(serialized_objs[3])
+            # TODO: fix volatiles under deserialization support for __setstate__ and __getstate__
+            self.add_volatiles_under_deserialization(serialized_objs.vol_objs.values())
             
             # Get EE
             try:
                 execution_client = self.ready_clients[location]
             except KeyError:
-                exeenv = self.get_execution_environments_info()[location] 
+                exeenv = self.get_execution_environment_info(location)
                 self.logger.debug("Not found in cache ExecutionEnvironment {%s}! Starting it at %s:%d",
                                    location, exeenv.hostname, exeenv.port)
                 execution_client = EEClient(exeenv.hostname, exeenv.port)
@@ -132,14 +141,20 @@ class ClientRuntime(DataClayRuntime):
     
             # Call EE
             self.logger.verbose("Calling make persistent to EE %s ", location)
-            execution_client.make_persistent(settings.current_session_id, serialized_objs)
+            execution_client.make_persistent(settings.current_session_id, serialized_objs.vol_objs.values())
 
             # update the hint with the location, and return it
             instance.set_hint(location)
             
             # remove volatiles under deserialization
-            self.remove_volatiles_under_deserialization()
-        
+            self.remove_volatiles_under_deserialization(serialized_objs.vol_objs.values())
+
+        object_id = instance.get_object_id()
+        locations = set()
+        locations.add(location)
+        metadata_info = MetaDataInfo(object_id, False, instance.get_dataset_id(),
+                                     instance.get_class_extradata().class_id, locations, alias, None)
+        self.metadata_cache[object_id] = metadata_info
         return location
     
     def add_session_reference(self, object_id):
@@ -155,13 +170,14 @@ class ClientRuntime(DataClayRuntime):
         object_id = instance.get_object_id()
         
         self.logger.verbose("Calling operation '%s' in object with ID %s", operation_name, object_id)
-        self.logger.debug("Call is being done into %r with #%d parameters",
-                          instance, len(parameters))
+        self.logger.debug("Call is being done into object with ID %s with #%d parameters",
+                          object_id, len(parameters))
         
         using_hint = False
         if instance.get_hint() is not None:
             exeenv_id = instance.get_hint()
             self.logger.verbose("Using hint = %s", exeenv_id)
+            using_hint = True
         else:
             exeenv_id = next(iter(self.get_metadata(object_id).locations))
             
@@ -177,13 +193,106 @@ class ClientRuntime(DataClayRuntime):
     def get_implementation_id(self, object_id, operation_name, implementation_idx=0):
         operation = self.get_operation_info(object_id, operation_name)
         return operation.remoteImplID
-    
+
+    def delete_alias(self, dc_obj):
+        session_id = self.get_session_id()
+        hint = dc_obj.get_hint()
+        object_id = dc_obj.get_object_id()
+        exec_location_id = hint
+        if exec_location_id is None:
+            exec_location_id = self.get_location(object_id)
+        try:
+            execution_client = self.ready_clients[exec_location_id]
+        except KeyError:
+            backend_to_call = self.get_execution_environment_info(exec_location_id)
+            execution_client = EEClient(backend_to_call.hostname, backend_to_call.port)
+            self.ready_clients[exec_location_id] = execution_client
+        execution_client.delete_alias(session_id, object_id)
+        alias = dc_obj.get_alias()
+        if alias is not None:
+            if alias in self.alias_cache:
+                del self.alias_cache[alias]
+        dc_obj.set_alias(None)
+
     def close_session(self):
-         self.ready_clients["@LM"].close_session(settings.current_session_id)
+        self.logger.debug("** Closing session **")
+        self.ready_clients["@LM"].close_session(settings.current_session_id)
 
     def get_hint(self):
         return None
     
     def get_session_id(self):
         return settings.current_session_id
-    
+
+    def synchronize(self, instance, operation_name, params):
+        session_id = self.get_session_id()
+        object_id = instance.get_object_id()
+        dest_backend_id = self.get_location(instance.get_object_id())
+        operation = self.get_operation_info(instance.get_object_id(), operation_name)
+        implementation_id = self.get_implementation_id(instance.get_object_id(), operation_name)
+        # === SERIALIZE PARAMETER ===
+        serialized_params = SerializationLibUtilsSingleton.serialize_params_or_return(
+            params=[params],
+            iface_bitmaps=None,
+            params_spec=operation.params,
+            params_order=operation.paramsOrder,
+            hint_volatiles=instance.get_hint(),
+            runtime=self)
+        try:
+            execution_client = self.ready_clients[dest_backend_id]
+        except KeyError:
+            exeenv = self.get_execution_environment_info(dest_backend_id)
+            execution_client = EEClient(exeenv.hostname, exeenv.port)
+            self.ready_clients[dest_backend_id] = execution_client
+        execution_client.synchronize(session_id, object_id, implementation_id, serialized_params)
+
+    def detach_object_from_session(self, object_id, hint):
+        try:
+            cur_session = self.get_session_id()
+            exec_location_id = hint
+            if exec_location_id is None:
+                exec_location_id = self.get_location(object_id)
+            try:
+                execution_client = self.ready_clients[exec_location_id]
+            except KeyError:
+                backend_to_call = self.get_execution_environment_info(exec_location_id)
+                execution_client = EEClient(backend_to_call.hostname, backend_to_call.port)
+                self.ready_clients[exec_location_id] = execution_client
+            execution_client.detach_object_from_session(object_id, cur_session)
+        except:
+            traceback.print_exc()
+
+    def federate_to_backend(self, dc_obj, external_execution_environment_id, recursive):
+        object_id = dc_obj.get_object_id()
+        hint = dc_obj.get_hint()
+        session_id = self.get_session_id()
+        exec_location_id = hint
+        if exec_location_id is None:
+            exec_location_id = self.get_location(object_id)
+        try:
+            execution_client = self.ready_clients[exec_location_id]
+        except KeyError:
+            exeenv = self.get_execution_environment_info(exec_location_id)
+            execution_client = EEClient(exeenv.hostname, exeenv.port)
+            self.ready_clients[exec_location_id] = execution_client
+
+        self.logger.debug("[==FederateObject==] Starting federation of object by %s calling EE %s with dest dataClay %s, and session %s",
+                          object_id, exec_location_id, external_execution_environment_id, session_id)
+        execution_client.federate(session_id, object_id, external_execution_environment_id, recursive)
+
+    def unfederate_from_backend(self, dc_obj, external_execution_environment_id, recursive):
+        object_id = dc_obj.get_object_id()
+        hint = dc_obj.get_hint()
+        session_id = self.get_session_id()
+        self.logger.debug("[==UnfederateObject==] Starting unfederation of object %s with ext backend %s, and session %s", object_id, external_execution_environment_id, session_id)
+        exec_location_id = hint
+        if exec_location_id is None:
+            exec_location_id = self.get_location(object_id)
+        try:
+            execution_client = self.ready_clients[exec_location_id]
+        except KeyError:
+            exeenv = self.get_execution_environment_info(exec_location_id)
+            execution_client = EEClient(exeenv.hostname, exeenv.port)
+            self.ready_clients[exec_location_id] = execution_client
+
+        execution_client.unfederate(session_id, object_id, external_execution_environment_id, recursive)
