@@ -18,6 +18,8 @@ from dataclay.util import Configuration
 __author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
 __copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'
 
+from dataclay.util.management.metadataservice.RegistrationInfo import RegistrationInfo
+
 logger = logging.getLogger(__name__)
 
 current_milli_time = lambda: int(round(time.time() * 1000))
@@ -52,14 +54,6 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
         Must be thread-safe.
         """
         self.session_expires_dates = dict()
-    
-        """
-        Alias references found. TODO: this set is not cleaned. Important: modify this when design of removeAlias: we need to add
-        'alias' as DataClayObject field since alias can be added dynamically to a volatile. Or check updateAliases function. Also,
-        ensure removeAlias in LM notifies EE and avoid race condition remove alias but continue using object and GlobalGC clean it.
-        Should be thread-safe.
-        """
-        self.alias_references = set()
 
     def initialize_runtime_aux(self):
         self.dataclay_heap_manager = ExecutionEnvironmentHeapManager(self)
@@ -102,7 +96,10 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
         @postcondition: Get Session ID associated to current thread 
         @return: Session ID associated to current thread 
         """
-        return threadLocal.session_id
+        if hasattr(threadLocal, "session_id"):
+            return threadLocal.session_id
+        else:
+            return None
     
     def get_execution_environment(self):
         """
@@ -154,15 +151,18 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
             # 3 - object was persisted with an alias and it must be already registered -> we add a new alias.
             if instance.is_pending_to_register():
                 # Use case 1
-                reg_info = [instance.get_object_id(), instance.get_class_extradata().class_id,
-                            self.get_session_id(), instance.get_dataset_id()]
+                reg_infos = list()
+                reg_info = RegistrationInfo(instance.get_object_id(), instance.get_class_extradata().class_id,
+                            self.get_session_id(), instance.get_dataset_id(), alias)
+                reg_infos.append(reg_info)
                 # TODO: Review if we use hint of the object or the hint of the runtime.
-                new_object_id = self.ready_clients["@LM"].register_object(reg_info, instance.get_hint(), alias, LANG_PYTHON)
+                new_object_ids = self.ready_clients["@LM"].register_objects(reg_infos, instance.get_hint(), LANG_PYTHON)
+                new_object_id = next(iter(new_object_ids))
                 self.update_object_id(instance, new_object_id)
             else:
                 # Use case 2 and 3 - add new alias
-                self.ready_clients["@LM"].add_alias(instance.get_object_id(), alias)
-            
+                instance.set_alias(alias)
+
         return instance.get_location()
 
     def execute_implementation_aux(self, operation_name, instance, parameters, exeenv_id=None):
@@ -229,7 +229,6 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
         The return dictionary is the same in both cases, but note that the update
         should not use the provided instance for updating metadata to the LM.
         """
-        logger.debug("Store procedure for instance %r", instance)
         client = self.ready_clients["@STORAGE"]
         
         pending_objs = [instance]
@@ -256,8 +255,8 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
                         continue
     
                     dcc_extradata = current_obj.get_class_extradata()
-                    infos = [object_id, dcc_extradata.class_id,
-                             self.execution_environment.thread_local_info.session_id, dataset_id]
+                    infos = RegistrationInfo(object_id, dcc_extradata.class_id,
+                             self.execution_environment.thread_local_info.session_id, dataset_id, None)
                     reg_infos.append(infos)
     
                 # This object will soon be persistent
@@ -288,8 +287,7 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
         
         if make_persistent:
             lm_client = self.ready_clients["@LM"]
-            lm_client.register_objects(reg_infos, settings.environment_id, None, None,
-                                           LANG_PYTHON)
+            lm_client.register_objects(reg_infos, settings.environment_id, LANG_PYTHON)
             client.ds_store_objects(self.execution_environment.thread_local_info.session_id, serialized_objs, False, None)
         else:
             client.ds_upsert_objects(self.execution_environment.thread_local_info.session_id, serialized_objs)
@@ -313,13 +311,18 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
         :type ifacebitmaps: dict 
         :rtype: boolean
         """
-        
+
         object_id = volatile_obj.get_object_id()
-        if hasattr(self.thread_local_info, "volatiles_under_deserialitzation"): 
+        if hasattr(self.thread_local_info, "volatiles_under_deserialitzation"):
             if self.thread_local_info.volatiles_under_deserialitzation is not None: 
-                for obj_with_data in  self.thread_local_info.volatiles_under_deserialitzation.values():
-                    curr_obj_id = obj_with_data[0]
+                for obj_with_data in self.thread_local_info.volatiles_under_deserialitzation:
+                    curr_obj_id = obj_with_data.object_id
                     if object_id == curr_obj_id:
+                        if hasattr(volatile_obj, "__setstate__"):
+                            # Return true like the object is already deserialized, this will allow
+                            # any volatile under deserialization with __setstate__ to be actually deserialized
+                            # TODO: check race conditions
+                            return True
                         # deserialize it 
                         metaclass_id = volatile_obj.get_class_extradata().class_id
                         hint = volatile_obj.get_hint()
@@ -328,12 +331,6 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
                         return True
         return False
 
-    def add_alias_reference(self, object_id):
-        """
-        @summary Add +1 reference due to a new alias.
-        @param object_id ID of object with alias
-        """
-        self.alias_references.add(object_id)
         
     def add_session_reference(self, object_id):
         """
@@ -381,7 +378,14 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
                 self.session_expires_dates[session_id] = expiration_date
             finally: 
                 self.unlock(session_id)
-        
+
+    def delete_alias(self, dc_obj):
+        alias = dc_obj.get_alias()
+        if alias is not None:
+            self.delete_alias_in_dataclay(alias)
+        dc_obj.set_alias(None)
+        dc_obj.set_dirty(True)
+
     def close_session_in_ee(self, session_id):
         """
         @summary Close session in EE. Subtract session references for GC.
@@ -398,12 +402,12 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
         @summary Get retained refs by this EE
         @return Retained refs (alias, sessions, ...)
         """
-        logger.debug("Getting retained references")
-        retained_refs = list() 
+        retained_refs = list()
         
         """ memory references """
         retained_refs.extend(self.dataclay_heap_manager.get_object_ids_retained())
-        logger.debug("Obtained retained references in memory: %s " % str(len(retained_refs)))
+        if len(retained_refs) > 0:
+            logger.debug("Obtained retained references in memory: %s " % str(len(retained_refs)))
 
         """ session references """ 
         now = datetime.datetime.now()
@@ -430,7 +434,6 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
                 """
                 session_expired = False
                 expired_date = self.session_expires_dates.get(cur_session)
-                print(expired_date)
                 if expired_date is not None and now > expired_date:
                     if cur_session in self.quarantine_sessions:
                         # Session is actually removed 
@@ -468,7 +471,6 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
                         logger.debug("Removing session reference for oid %s" % str(oid))
                         for oid, oid_ref in self.references_hold_by_sessions.items():
                             oid_ref.discard(cur_session)  # do not raise key error if not found
-                        print(self.references_hold_by_sessions)
 
                 else:
                     retained_refs.append(oid)
@@ -486,7 +488,6 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
             if not obj_using_session:
                 del self.session_expires_dates[session_to_close]
         
-        logger.debug("Retained references %s" % str(retained_refs))
         return retained_refs
 
     def get_from_sl(self, object_id):
@@ -510,3 +511,40 @@ class ExecutionEnvironmentRuntime(DataClayRuntime):
         :rtype: None
         """
         return self.ready_clients["@STORAGE"].update_to_db(settings.environment_id, object_id, obj_bytes, dirty)
+
+    def synchronize(self, instance, operation_name, params):
+        session_id = self.get_session_id()
+        object_id = instance.get_object_id()
+        operation = self.get_operation_info(instance.get_object_id(), operation_name)
+        implementation_id = self.get_implementation_id(instance.get_object_id(), operation_name)
+        # === SERIALIZE PARAMETERS ===
+        serialized_params = SerializationLibUtilsSingleton.serialize_params_or_return(
+            params=[params],
+            iface_bitmaps=None,
+            params_spec=operation.params,
+            params_order=operation.paramsOrder,
+            hint_volatiles=instance.get_hint(),
+            runtime=self)
+        self.execution_environment.synchronize(session_id, object_id, implementation_id,
+                                               serialized_params)
+
+    def detach_object_from_session(self, object_id, hint):
+        cur_session = self.get_session_id()
+        if object_id in self.references_hold_by_sessions:
+            sessions_of_obj = self.references_hold_by_sessions.get(object_id)
+            if cur_session in sessions_of_obj:
+                sessions_of_obj.remove(cur_session)
+                logger.debug("Session %s removed from object %s" % (str(cur_session), str(object_id)))
+
+    def federate_to_backend(self, dc_obj, external_execution_environment_id, recursive):
+        object_id = dc_obj.get_object_id()
+        session_id = self.get_session_id()
+        self.logger.debug("[==FederateObject==] Starting federation of object by %s with dest dataClay %s, and session %s",
+                          object_id, external_execution_environment_id, session_id)
+        self.execution_environment.federate(session_id, object_id, external_execution_environment_id, recursive)
+
+    def unfederate_from_backend(self, dc_obj, external_execution_environment_id, recursive):
+        object_id = dc_obj.get_object_id()
+        session_id = self.get_session_id()
+        self.logger.debug("[==UnfederateObject==] Starting unfederation of object %s with ext backend %s, and session %s", object_id, external_execution_environment_id, session_id)
+        self.execution_environment.unfederate(session_id, object_id, external_execution_environment_id, recursive)
